/*------------------------------------------------------------------------------
 bin/djv_convert/djv_convert.cxx

 Copyright (c) 2004-2007 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

//#define DJV_DBG

#include <djv_base/user.h>
#include <djv_base/file.h>
#include <djv_base/sys.h>
#include <djv_base/box.h>
#include <djv_image/time.h>
#include <djv_image/io.h>
#include <djv_image/app.h>

using namespace djv;

class App : public image::app::App {
public:

  App(int, char **);

private:

  void work() throw (Error);

  Str time_label(int, const image::time::Speed &) const;
  Str time_label(double) const;
  Str info_label(const image::io::Info & in) const;
  
  bool cmdln(int, char **) throw (Error);
  Str help() const;

  File _in, _out;
  image::gl::data::Options _options;
  Str _load, _save;
  Str _time_start, _time_end;
  int _timeout;
  std::auto_ptr<image::pixel::PIXEL> _pixel;
  std::auto_ptr<image::time::FPS> _speed;
  image::tag::Tag _tag;
  bool _tag_auto;
  bool _timecode;
};

App::App(int argc, char ** argv) :
  image::app::App("djv_convert", argc, argv),
  _timeout(0), _tag_auto(true), _timecode(true)
{
  DJV_DBG_OBJ("App::App");
  
  // Command line.
  
  bool r = false;
  try { r = cmdln(argc, argv); }
  catch (Error in) { error(str::Format("Command Line: %1").arg(in)); }
  if (! r) { exit(1); return; }

  if (_in.seq_valid()) _in.type = file::SEQ;
  if (_out.seq_valid()) _out.type = file::SEQ;
  
  // Work.

  try { work(); }
  catch (Error error) { this->error(error); exit(1); }
}

void App::work() throw (Error)
{
  DJV_DBG_OBJ("App::work");
  DJV_DBG_PRN("input = " << _in);
  DJV_DBG_PRN("output = " << _out);

  djv::time::Timer timer;
  timer.start();
  
  // Open input.

  std::auto_ptr<image::io::Load> load;
  image::io::Info load_info;

  Error error;
  while (! load.get()) {
    try {
      load = std::auto_ptr<image::io::Load>(
        image::io::load()->get(_in, &load_info, _load));
    }
    catch (Error in) { error = in; }
    if (! load.get() && _timeout) {
      print("Timeout...");
      djv::time::sleep(1);
    } else
      break;
  }
  if (! load.get())
    throw Error(str::Format("%1: %2").arg(error).arg(_in.name()));

  if (_time_start.size())
    load_info.time.start = _timecode ?
      image::time::timecode_to_frame(
        image::time::str_to_timecode(_time_start), load_info.time.speed) :
      str::str_to_int(_time_start);
  if (_time_end.size())
    load_info.time.end = _timecode ?
      image::time::timecode_to_frame(
        image::time::str_to_timecode(_time_end), load_info.time.speed) :
      str::str_to_int(_time_end);
  
  print(str::Format("%1 %2").arg(_in).arg(info_label(load_info)));
  
  // Open output.

  image::io::Info save_info;
  save_info.size = size_valid(_options.scale) ? _options.scale : load_info.size;
  save_info.pixel = _pixel.get() ? *_pixel : load_info.pixel;
  save_info.tag = load_info.tag;
  save_info.tag.add(_tag);
  
  if (_out.seq_valid()) {
    save_info.time.start = save_info.time.end = _out.seq.start;
    if (_out.seq.start != _out.seq.end)
      save_info.time.end = _out.seq.end;
  }
  if (save_info.time.start == save_info.time.end)
    save_info.time.end = save_info.time.start +
      load_info.time.end - load_info.time.start;
  save_info.time.speed = _speed.get() ? *_speed : load_info.time.speed;

  std::auto_ptr<image::io::Save> save;
  try {
    save = std::auto_ptr<image::io::Save>(
      image::io::save()->get(_out, save_info, _save));
  }
  catch (Error in)
    { throw Error(str::Format("%1: %2").arg(_out.name()).arg(in)); }

  print(str::Format("%1 %2").arg(_out).arg(info_label(save_info)));

  // Work.

  const int length = save_info.time.end - save_info.time.start + 1;
  const int progress_length = 10;
  int progress_count = 0;
  double frame_time = 0.0;
  for (int i = 0; i < length; ++i) {

    djv::time::Timer frame_timer;
    frame_timer.start();
    
    // Load.

    const image::Image * image = 0;
    
    int timeout = _timeout;
    while (! image) {
      try { image = load->get(load_info.time.start + i); }
      catch (Error in) { error = in; }
      if (! image && timeout > 0) {
        //print("Timeout...");
        --timeout;
        djv::time::sleep(1);
      } else
        break;
    }
    if (! image)
      throw Error(str::Format("%1: %2").arg(error).arg(_in.name()));
    DJV_DBG_PRN("image = " << *image);

    // Image tags.
    
    image::tag::Tag tag = image->info().tag;
    if (_tag_auto) {
      tag[image::tag::label()[image::tag::CREATOR]] = user::current();
      tag[image::tag::label()[image::tag::TIME]] =
        djv::time::time_to_str(djv::time::current());
      tag[image::tag::label()[image::tag::TIMECODE]] =
        image::time::timecode_to_str(image::time::frame_to_timecode(
          save_info.time.start + i, save_info.time.speed));
    }
    tag.add(_tag);
    save_info.tag = tag;
    
    image::Image tmp(save_info);
    DJV_DBG_PRN("tmp = " << tmp);
    
    // Process.
    
    if (! size_valid(_options.scale)) _options.scale = save_info.size;
    _options.profile = image->info().profile;
    image::gl::data::copy(*image, &tmp, _options);
    
    // Save.

    DJV_DBG_PRN("output = " << tmp);
    try { save->set(tmp, save_info.time.start + i); }
    catch (Error in)
      { throw Error(str::Format("%1: %2").arg(_out.name()).arg(in)); }

    // Statistics.

    frame_timer.check();
    frame_time += frame_timer.seconds();

    if (length > 1)
      if (i / double(length) * progress_length > progress_count) {
        print(str::Format("[%1%] Estimated Time = %2").
          arg(int(i / double(length) * 100.0), 3).
          arg(time_label(frame_time / double(i + 1) * (length - (i + 1)))));
        ++progress_count;
      }
  }
  if (length > 1) print("[100%] ", false);
  
  try { save->close(); }
  catch (Error in)
    { throw Error(str::Format("%1: %2").arg(_out.name()).arg(in)); }

  timer.check();
  print(str::Format("Time Elapsed = %1 (%2 frames/second)").
    arg(time_label(timer.seconds())).arg(length / timer.seconds(), 0, -2));
}

Str App::time_label(int in, const image::time::Speed & speed) const
{
  return _timecode ?
    image::time::timecode_to_str(image::time::frame_to_timecode(in, speed)) :
    str::int_to_str(in);
}

Str App::time_label(double in) const
{
  int hour, minute;
  double second;
  djv::time::seconds_to_time(in, &hour, &minute, &second);
  return str::Format("%1:%2:%3").
    arg(hour, 2, '0').arg(minute, 2, '0').arg(second, 5, 2, '0');
}

Str App::info_label(const image::io::Info & in) const
{
  return str::Format("%1x%2:%3 %4 %5-%6@%7").
    arg(in.size.x).arg(in.size.y).arg(aspect(in.size), 0, -2).
    arg(image::pixel::pixel_label()[in.pixel]).
    arg(time_label(in.time.start, in.time.speed)).
    arg(time_label(in.time.end, in.time.speed)).
    arg(image::time::speed_to_float(in.time.speed));
}

bool App::cmdln(int argc, char ** argv) throw (Error)
{
  if (1 == argc) { message(help()); return false; }
  if (! app::App::cmdln(argc, argv)) return false;
  
  List<Str> list, arg_list;
  for (int i = 1; i < argc; ++i) list += argv[i];
  
  while (list.size()) {
    const Str arg = list.pop_front();
    
    // Options.
    
    if (str::compare("-x_timecode", arg) || str::compare("-t", arg)) {
      _timecode = false;
    }
    
    // Image options.

    else if (str::compare("-mirror_x", arg)) {
      _options.mirror_x = true;
    } else if (str::compare("-mirror_y", arg)) {
      _options.mirror_y = true;
    } else if (str::compare("-scale", arg)) {
      if (! vec::serialize(&list, &_options.scale)) throw Error(arg);
    } else if (str::compare("-channel", arg)) {
      if (! str::serialize(&list, &_options.channel,
        image::gl::data::channel_label())) throw Error(arg);
    } else if (str::compare("-filter", arg)) {
      if (! str::serialize(&list, &_options.filter,
        image::gl::data::filter_label())) throw Error(arg);
    }
    
    // Load options.

    else if (str::compare("-load", arg)) {
      Str plugin, option;
      if (! str::serialize(&list, &plugin)) throw Error(arg);
      if (! str::serialize(&list, &option)) throw Error(arg);
      if (! image::io::load()->option(plugin, option, &list)) throw Error(arg);
    } else if (str::compare("-load_plugin", arg)) {
      if (! str::serialize(&list, &_load)) throw Error(arg);
    } else if (str::compare("-time", arg)) {
      if (! str::serialize(&list, &_time_start)) throw Error(arg);
      if (! str::serialize(&list, &_time_end)) throw Error(arg);
    } else if (str::compare("-timeout", arg)) {
      if (! str::serialize(&list, &_timeout)) throw Error(arg);
    }
    
    // Save options.
    
    else if (str::compare("-save", arg)) {
      Str plugin, option;
      if (! str::serialize(&list, &plugin)) throw Error(arg);
      if (! str::serialize(&list, &option)) throw Error(arg);
      if (! image::io::save()->option(plugin, option, &list)) throw Error(arg);
    } else if (str::compare("-save_plugin", arg)) {
      if (! str::serialize(&list, &_save)) throw Error(arg);
    } else if (str::compare("-pixel", arg) || str::compare("-p", arg)) {
      image::pixel::PIXEL value = image::pixel::PIXEL(0);
      if (! str::serialize(&list, &value, image::pixel::pixel_label()))
        throw Error(arg);
      _pixel = std::auto_ptr<image::pixel::PIXEL>(
        new image::pixel::PIXEL(value));
    } else if (str::compare("-speed", arg)) {
      image::time::FPS value = image::time::FPS(0);
      if (! str::serialize(&list, &value, image::time::fps_label()))
        throw Error(arg);
      _speed = std::auto_ptr<image::time::FPS>(new image::time::FPS(value));
    } else if (str::compare("-tag", arg)) {
      Str name, value;
      if (! str::serialize(&list, &name)) throw Error(arg);
      if (! str::serialize(&list, &value)) throw Error(arg);
      _tag[name] = value;
    } else if (str::compare("-x_tag", arg)) _tag_auto = false;

    // Arguments.
    
    else arg_list += arg;
  }
  
  if (arg_list.size() < 1) throw Error("input");
  _in = arg_list[0];
  if (arg_list.size() < 2) throw Error("output");
  _out = arg_list[1];
  if (arg_list.size() > 2) throw Error(arg_list[2]);
  
  return true;
}

Str App::help() const
{
  return str::Format(
"\n"
"djv_convert\n"
"-----------\n"
"Image conversion utility.\n"
"\n"
"Usage\n"
"-----\n"
"djv_convert (input) (output) [option]...\n"
"\n"
"Options\n"
"-------\n"
"-x_timecode, -t - Disable timecode display.\n"
"\n"
"Image Options\n"
"-------------\n"
"-mirror_x - Mirror image horizontally.\n"
"-mirror_y - Mirror image vertically.\n"
"\n"
"-scale (width) (height) - Scale image.\n"
"\n"
"-channel (value) - Channel: %1.\n"
"\n"
"-filter (value) - Filter: %2. Default: %3.\n"
"\n"
"Load Options\n"
"------------\n"
"-load (plugin) (option) (value) - Set a plugin option:\n"
"%4"
"\n"
"-load_plugin (value) - Specify load plugin: %5.\n"
"\n"
"-time (start) (end) - Start and end frames.\n"
"\n"
"-timeout (value) - Maximum number of seconds to wait for each input frame. "
"Default: %6.\n"
"\n"
" Save Options\n"
" ------------\n"
"-save (plugin) (option) (value) - Set a plugin option:\n"
"%7"
"\n"
"-save_plugin (value) - Specify save plugin: %8.\n"
"\n"
"-pixel, -p (value) - Pixel type: %9.\n"
"\n"
"-speed (value) - Speed: %10.\n"
"\n"
"-tag (name) (value) - Set an image tag.\n"
"-x_tag - Disable automatic image tag generation.\n"
"\n"
"%11"
"\n"
"Examples\n"
"--------\n"
"> djv_convert input.sgi output.tga\n"
"Convert an image.\n"
"\n"
"> djv_convert input.1-100.sgi output.1.tga\n"
"Convert an image sequence. Only the first frame of the output needs to be "
"specified.\n"
"\n"
"> djv_convert input.mov output.1.tga\n"
"Convert a movie to an image sequence.\n"
"\n"
"> djv_convert input.0001-0100.dpx output.mov\n"
"Convert an image sequence to a movie.\n"
"\n"
"> djv_convert input.sgi output.sgi -pixel \"rgb u16\"\n"
"Convert pixel type.\n"
"\n"
"> djv_convert input.sgi output.sgi -scale 2048 1556\n"
"Scale an image.\n"
"\n"
"> djv_convert $DISPLAY lut.1dl -load_plugin vlut\n"
"Save video lookup table to a file.\n"
"\n"
"> djv_convert lut.1dl $DISPLAY -save_plugin vlut\n"
"Load video lookup table from a file.\n"
  ).
    arg(str::join(image::gl::data::channel_label(), ", ")).
    arg(str::join(image::gl::data::filter_label(), ", ")).
    arg(image::gl::data::filter_label()[_options.filter]).
    arg(image::io::load()->help()).
    arg(str::join(image::io::load()->name_list(), ", ")).
    arg(_timeout).
    arg(image::io::save()->help()).
    arg(str::join(image::io::save()->name_list(), ", ")).
    arg(str::join(image::pixel::pixel_label(), ", ")).
    arg(str::join(image::time::fps_label(), ", ")).
    arg(app::App::help());
}

int main(int argc, char ** argv)
{
  int r = 1;
  try { r = App(argc, argv).exec(); }
  catch (Error in) { error::print(in); }
  return r;
}

