/*------------------------------------------------------------------------------
 bin/djv_info/djv_info.cxx

 Copyright (c) 2004-2007 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

//#define DJV_DBG

#include <djv_base/sys.h>
#include <djv_base/file.h>
#include <djv_base/dir.h>
#include <djv_image/time.h>
#include <djv_image/io.h>
#include <djv_image/app.h>

using namespace djv;

class App : public image::app::App {
public:

  App(int, char **);

private:

  void file(const File &, bool path = false) throw (Error);
  void dir(const File &, bool label) throw (Error);

  Str time_label(int, const image::time::Speed &) const;

  bool cmdln(int, char **) throw (Error);
  Str help() const;

  List<Str> _input;
  bool _verbose, _seq, _recurse, _timecode;
  int _columns;
  Str _load;
};

App::App(int argc, char ** argv) :
  image::app::App("djv_info", argc, argv),
  _verbose(false), _seq(true), _recurse(false), _timecode(true),
  _columns(sys::terminal_width())
{
  DJV_DBG_OBJ("App::App");
    
  // Command line.
  
  bool r = false;
  try { r = cmdln(argc, argv); }
  catch (Error in) { error(str::Format("Command Line: %1").arg(in)); }
  if (! r) { exit(1); return; }

  // Setup.

  if (! _input.size()) _input += File(".");
  DJV_DBG_PRN("input = " << _input);
  
  List<File> list;
  for (uint i = 0; i < _input.size(); ++i) {
    File file(_input[i]);
    if (_seq && file.seq_valid()) {
      file.type = file::SEQ;
      for (int j = file.seq.start; j <= file.seq.end; ++j)
        list += file.get(j);
    } else {
      list += file;
    }
  }
  if (_seq) file::seq_compress(&list);
  DJV_DBG_PRN("files = " << list);
  
  // Work.
  
  for (uint i = 0; i < list.size(); ++i)
    if (file::DIR == list[i].type) {
      try { dir(list[i], (list.size() > 1) || _recurse); }
      catch (Error in) { error(in); r = false; }
    } else {
      try { file(list[i]); }
      catch (Error in) { error(in); r = false; }
    }
  exit(! r);
}

void App::file(const File & in, bool path) throw (Error)
{
  DJV_DBG_OBJ("App::file");
  DJV_DBG_PRN("in = " << in);
  
  // Open.

  image::io::Info info;
  try { std::auto_ptr<image::io::Load>(image::io::load(in, &info, _load)); }
  catch (Error error)
    { throw Error(str::Format("%1: %2").arg(error).arg(in.name())); }
  
  // Print.

  Str name = in.get(-1, path);
  
  if (! _verbose) {
  
    const Str info_str = str::Format("%1x%2:%3 %4 %5@%6").
      arg(info.size.x).arg(info.size.y).arg(aspect(info.size), 0, -2).
      arg(image::pixel::pixel_label()[info.pixel]).
      arg(time_label(info.time.end - info.time.start + 1, info.time.speed)).
      arg(image::time::speed_to_float(info.time.speed), 0, -2);

    if (_columns) {
      const uint size = math::max(_columns - int(info_str.size()) - 2, 0);
      if (name.size() > size) {
        name = std::string(name, 0, size);
        size_t i = name.size(), j = 3;
        while (i-- > 0 && j-- > 0) name[i] = '.';
      }
    }

    print(str::Format("%1 %2").arg(name).
      arg(info_str, _columns - int(name.size()) - 2));
    
  } else {
  
    print(name);

    print(str::Format("  Width = %1").arg(info.size.x));
    print(str::Format("  Height = %1").arg(info.size.y));
    print(str::Format("  Aspect = %1").arg(aspect(info.size), 0, -2));
    print(str::Format("  Pixel = %1").
      arg(image::pixel::pixel_label()[info.pixel]));
    print(str::Format("  Start = %1").
      arg(time_label(info.time.start, info.time.speed)));
    print(str::Format("  End = %1").
      arg(time_label(info.time.end, info.time.speed)));
    print(str::Format("  Duration = %1").
      arg(time_label(info.time.end - info.time.start + 1, info.time.speed)));
    print(str::Format("  Speed = %1").
      arg(image::time::speed_to_float(info.time.speed)));
    
    const List<Str> key_list = info.tag.key_list();
    for (uint i = 0; i < key_list.size(); ++i)
      print(str::Format("  Tag %1 = %2").arg(key_list[i]).
        arg(info.tag[key_list[i]]));

    separator();
  }
}

void App::dir(const File & in, bool label) throw (Error)
{
  DJV_DBG_OBJ("App::dir");
  DJV_DBG_PRN("in = " << in);
  
  // Setup.

  Dir dir(in);
  if (! dir.valid())
    throw Error(str::Format("%1: Open Error").arg(in.name()));

  List<File> list = dir.list();
  if (_seq) file::seq_compress(&list);

  // Print.

  if (label) {
    print(str::Format("%1:").arg(in));
    if (_verbose) separator();
  }
  
  for (uint i = 0; i < list.size(); ++i) {
    try { file(list[i]); }
    catch (Error) {}
  }
  
  separator();
  
  // Recurse.
  
  if (_recurse) {
    Dir dir(in);
    List<File> list = dir.list();
    file::filter(&list, file::FILTER_FILE | file::FILTER_HIDDEN);
    for (uint i = 0; i < list.size(); ++i)
      this->dir(list[i], label);
  }
}

Str App::time_label(int in, const image::time::Speed & speed) const
{
  return _timecode ?
    image::time::timecode_to_str(image::time::frame_to_timecode(in, speed)) :
    str::int_to_str(in);
}

bool App::cmdln(int argc, char ** argv) throw (Error)
{
  if (1 == argc) { message(help()); return false; }
  if (! app::App::cmdln(argc, argv)) return false;
  
  List<Str> list;
  for (int i = 1; i < argc; ++i) list += argv[i];
  
  while (list.size()) {
    const Str arg = list.pop_front();
    
    // Options.
    
    if (str::compare("-verbose", arg) || str::compare("-v", arg)) {
      _verbose = true;
    } else if (str::compare("-x_seq", arg) || str::compare("-q", arg)) {
      _seq = false;
    } else if (str::compare("-recurse", arg) || str::compare("-r", arg)) {
      _recurse = true;
    } else if (str::compare("-x_timecode", arg) || str::compare("-t", arg)) {
      _timecode = false;
    } else if (str::compare("-columns", arg) || str::compare("-c", arg)) {
      if (! str::serialize(&list, &_columns)) throw Error(arg);
    }
    
    // Load options.
    
    else if (str::compare("-load", arg)) {
      Str plugin, option;
      if (! str::serialize(&list, &plugin)) throw Error(arg);
      if (! str::serialize(&list, &option)) throw Error(arg);
      if (! image::io::load()->option(plugin, option, &list)) throw Error(arg);
    } else if (str::compare("-load_plugin", arg)) {
      if (! str::serialize(&list, &_load)) throw Error(arg);
    }
    
    // Arguments.
    
    else _input += arg;
  }

  if (! _input.size()) throw Error("input");
  
  return true;
}

Str App::help() const
{
  return str::Format(
"\n"
"djv_info\n"
"--------\n"
"Image information utility.\n"
"\n"
"Output:\n"
"\n"
"yesterdayview.mov                         640x424:1.51 RGB U8 00:02:00:01@12\n"
"0001-1000.tga                              720x480:1.5 RGB U8 00:00:41:16@24\n"
"dlad.dpx                               2048x1556:1.32 RGB U10 00:00:00:01@24\n"
"\n"
"(name) (width)x(height):(aspect) (format) (type) (duration):(speed)\n"
"\n"
"Usage\n"
"-----\n"
"djv_info (image|directory)... [option]...\n"
"\n"
"Options\n"
"-------\n"
"-verbose, -v - Verbose information.\n"
"-x_seq, -q - Disable file sequencing.\n"
"-recurse, -r - Descend into sub-directories.\n"
"-x_timecode, -t - Disable timecode display.\n"
"-columns, -c (value) - Number of output columns. Zero disables formatting.\n"
"\n"
"Load Options\n"
"------------\n"
"-load (plugin) (option) (value) - Set a plugin option:\n"
"%1"
"\n"
"-load_plugin (value) - Specify load plugin: %2.\n"
"\n"
"%3"
"\n"
"Examples\n"
"--------\n"
"> djv_info image.sgi image2.sgi\n"
"Display image information.\n"
"\n"
"> djv_info image.1-100.sgi\n"
"Display image sequence information.\n"
"\n"
"> djv_info ~/pics\n"
"Display information about all images within a directory.\n"
"\n"
"> djv_info $DISPLAY -load_plugin vlut\n"
"Display video lookup table information.\n"
  ).
    arg(image::io::load()->help()).
    arg(str::join(image::io::load()->name_list(), ", ")).
    arg(app::App::help());
}

int main(int argc, char ** argv)
{
  int r = 1;
  try { r = App(argc, argv).exec(); }
  catch (Error in) { error::print(in); }
  return r;
}

