/*------------------------------------------------------------------------------
 lib/djv_image/gl_data_fnc.cxx

 Copyright (c) 2004-2007 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

#include "gl_data.h"
#include <djv_base/mtx.h>

namespace djv {
namespace image {
namespace gl {
namespace data {

Mtx4<GLfloat> color_mtx(const Display::Color & in)
{
  const Mtx4<GLfloat> brightness(
    in.brightness, 0, 0, 0,
    0, in.brightness, 0, 0,
    0, 0, in.brightness, 0,
    0, 0, 0, 1);

  const Mtx4<GLfloat> contrast =
    Mtx4<GLfloat>(
      1, 0, 0, 0,
      0, 1, 0, 0,
      0, 0, 1, 0,
      -0.5, -0.5, -0.5, 1) *
    Mtx4<GLfloat>(
      in.contrast, 0, 0, 0,
      0, in.contrast, 0, 0,
      0, 0, in.contrast, 0,
      0, 0, 0, 1) *
    Mtx4<GLfloat>(
      1, 0, 0, 0,
      0, 1, 0, 0,
      0, 0, 1, 0,
      0.5, 0.5, 0.5, 1);

  const float s [] = {
    (1.0 - in.saturation) * 0.3086,
    (1.0 - in.saturation) * 0.6094,
    (1.0 - in.saturation) * 0.0820 };
  const Mtx4<GLfloat> saturation(
    s[0] + in.saturation, s[0], s[0], 0,
    s[1], s[1] + in.saturation, s[1], 0,
    s[2], s[2], s[2] + in.saturation, 0,
    0, 0, 0, 1 );

  return brightness * contrast * saturation;
}

image::Data color_levels(const Display::Levels & in, double soft_clip)
{
  image::Data out(image::data::Info(Vec2i(1024, 1), pixel::L_F32));
  
  const double in_tmp = in.in[1] - in.in[0];
  const double gamma = 1.0 / in.gamma;
  const double out_tmp = in.out[1] - in.out[0];

  pixel::F32_T * p = reinterpret_cast<pixel::F32_T *>(out.data());
  const int size = out.size().x;
  for (int i = 0; i < size; ++i, ++p)
    *p = math::soft_clip(
           (math::pow(math::max(
             ((i / double(size - 1) - in.in[0]) / in_tmp), 0.000001),
             gamma) * out_tmp + in.out[0]),
           soft_clip);
  
  return out;
}

bool operator == (const Display & a, const Display & b)
{
  return a.lut.info() == b.lut.info() && a.color == b.color &&
    a.levels == b.levels && a.soft_clip == b.soft_clip;
}

bool operator == (const Display::Color & a, const Display::Color & b)
{
  return a.brightness == b.brightness && a.contrast == b.contrast &&
    a.saturation == b.saturation;
}

bool operator == (const Display::Levels & a, const Display::Levels & b)
{
  return a.in[0] == b.in[0] && a.in[1] == b.in[1] && a.gamma == b.gamma &&
    a.out[0] == b.out[0] && a.out[1] == b.out[1];
}

bool operator != (const Display & a, const Display & b)
  { return ! (a == b); }
bool operator != (const Display::Color & a, const Display::Color & b)
  { return ! (a == b); }
bool operator != (const Display::Levels & a, const Display::Levels & b)
  { return ! (a == b); }

bool serialize(List<Str> * in, Display * out)
{
  if (! serialize(in, &out->color)) return false;
  if (! serialize(in, &out->levels)) return false;
  if (! str::serialize(in, &out->soft_clip)) return false;
  return true;
}

List<Str> serialize(const Display & in)
{
  return List<Str>() << serialize(in.color) << serialize(in.levels) <<
    str::serialize(in.soft_clip);
}

bool serialize(List<Str> * in, Display::Color * out)
{
  if (! str::serialize(in, &out->brightness)) return false;
  if (! str::serialize(in, &out->contrast)) return false;
  if (! str::serialize(in, &out->saturation)) return false;
  return true;
}

List<Str> serialize(const Display::Color & in)
{
  return List<Str>() << str::serialize(in.brightness) <<
    str::serialize(in.contrast) << str::serialize(in.saturation);
}

bool serialize(List<Str> * in, Display::Levels * out)
{
  if (! str::serialize(in, &out->in[0])) return false;
  if (! str::serialize(in, &out->in[1])) return false;
  if (! str::serialize(in, &out->gamma)) return false;
  if (! str::serialize(in, &out->out[0])) return false;
  if (! str::serialize(in, &out->out[1])) return false;
  return true;
}

List<Str> serialize(const Display::Levels & in)
{
  return List<Str>() << str::serialize(in.in[0]) << str::serialize(in.in[1]) <<
    str::serialize(in.gamma) << str::serialize(in.out[0]) <<
    str::serialize(in.out[1]);
}

const List<Str> & channel_label()
{
  static const List<Str> data = List<Str>() << "Default" << "Red" << "Green" <<
    "Blue" << "Alpha";
  return data;
}

const List<Str> & filter_label()
{
  static const List<Str> data = List<Str>() << "Linear" << "Nearest";
  return data;
}

Dbg & operator << (Dbg & dbg, const Display & in)
  { return dbg << serialize(in.color) << " " << serialize(in.levels) << " " <<
      in.soft_clip; }
Dbg & operator << (Dbg & dbg, CHANNEL in)
  { return dbg << channel_label()[in]; }
Dbg & operator << (Dbg & dbg, FILTER in)
  { return dbg << filter_label()[in]; }

}
}
}
}

