/*------------------------------------------------------------------------------
 lib/djv_image/gl_data_shader.cxx

 Copyright (c) 2004-2007 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

//#define DJV_DBG
#define _DBG "lib/djv_image/gl_data_shader.cxx"

#include "_gl_data.h"
#include <djv_base/mtx.h>

namespace djv {
namespace image {
namespace gl {
namespace data {

namespace {

const Str
  error_namespace = "OpenGL",
  error_texture_max = "Maximum Texture Size: %1",
  error_shader_create = "Create Shader",
  error_shader_compile = "Compile Shader:\n%1",
  error_shader_program = "Create Shader Program",
  error_shader_link = "Link Shader:\n%1";

const Str shader_source =
"#version 110\n"
"#extension GL_ARB_texture_rectangle : enable\n"
"\n"
"uniform sampler2DRect texture;\n"
"\n"
"uniform struct {\n"
"  sampler1D lut;\n"
"  struct { float v, d, k, f, g; } exposure;\n"
"} profile;\n"
"\n"
"uniform struct {\n"
"  sampler1D lut;\n"
"  mat4 color;\n"
"  float levels_in, levels_in_tmp, gamma, levels_out, levels_out_tmp;\n"
"  float soft_clip;\n"
"} display;\n"
"\n"
"float knee(const float value, const float f)\n"
"{\n"
"  return log(value * f + 1.0) / f;\n"
"}\n"
"\n"
"vec4 soft_clip(vec4 value, const float soft_clip)\n"
"{\n"
"  const float tmp = 1.0 - soft_clip;\n"
"  if (value[0] > tmp)\n"
"    value[0] = tmp + (1.0 - exp(-(value[0] - tmp) / soft_clip)) * soft_clip;\n"
"  if (value[1] > tmp)\n"
"    value[1] = tmp + (1.0 - exp(-(value[1] - tmp) / soft_clip)) * soft_clip;\n"
"  if (value[2] > tmp)\n"
"    value[2] = tmp + (1.0 - exp(-(value[2] - tmp) / soft_clip)) * soft_clip;\n"
"  return value;\n"
"}\n"
"\n"
"void main(void)\n"
"{\n"
"  vec4 color = texture2DRect(texture, gl_TexCoord[0].st);\n"
"%1"
"  gl_FragColor = color;\n"
"}\n";

const Str profile_lut_source =
"color[0] = texture1D(profile.lut, color[0])[0];\n"
"color[1] = texture1D(profile.lut, color[1])[1];\n"
"color[2] = texture1D(profile.lut, color[2])[2];\n"
"color[3] = texture1D(profile.lut, color[3])[3];\n";

const Str exposure_source =
"color[0] = max(0.0, color[0] - profile.exposure.d) * profile.exposure.v;\n"
"color[1] = max(0.0, color[1] - profile.exposure.d) * profile.exposure.v;\n"
"color[2] = max(0.0, color[2] - profile.exposure.d) * profile.exposure.v;\n"
"\n"
"if (color[0] > profile.exposure.k)\n"
"  color[0] = profile.exposure.k + knee(color[0] - profile.exposure.k,\n"
"    profile.exposure.f);\n"
"if (color[1] > profile.exposure.k)\n"
"  color[1] = profile.exposure.k + knee(color[1] - profile.exposure.k,\n"
"    profile.exposure.f);\n"
"if (color[2] > profile.exposure.k)\n"
"  color[2] = profile.exposure.k + knee(color[2] - profile.exposure.k,\n"
"    profile.exposure.f);\n"
"\n"
"color[0] = pow(color[0], profile.exposure.g) * 0.332;\n"
"color[1] = pow(color[1], profile.exposure.g) * 0.332;\n"
"color[2] = pow(color[2], profile.exposure.g) * 0.332;\n";

const Str display_lut_source =
"color[0] = texture1D(display.lut, color[0])[0];\n"
"color[1] = texture1D(display.lut, color[1])[1];\n"
"color[2] = texture1D(display.lut, color[2])[2];\n"
"color[3] = texture1D(display.lut, color[3])[3];\n";

const Str display_color_source =
"color *= display.color;\n";

const Str display_levels_source =
"color[0] = pow(max(color[0] - display.levels_in, 0.0) /\n"
"  display.levels_in_tmp, display.gamma) * display.levels_out_tmp +\n"
"  display.levels_out;\n"
"color[1] = pow(max(color[1] - display.levels_in, 0.0) /\n"
"  display.levels_in_tmp, display.gamma) * display.levels_out_tmp +\n"
"  display.levels_out;\n"
"color[2] = pow(max(color[2] - display.levels_in, 0.0) /\n"
"  display.levels_in_tmp, display.gamma) * display.levels_out_tmp +\n"
"  display.levels_out;\n";

const Str display_soft_clip_source =
"color = soft_clip(color, display.soft_clip);\n";

const Str channel_source =
"color = vec4(color[%1]);\n";

Str source(profile::PROFILE profile, const Display & display, CHANNEL channel)
{
  Str tmp;
  
  switch (profile) {
    case profile::LUT: tmp += profile_lut_source; break;
    case profile::EXPOSURE: tmp += exposure_source; break;
    default: break;
  }
  
  if (vec::size_valid(display.lut.size()))
    tmp += display_lut_source;
  if (display.color != Display().color)
    tmp += display_color_source;
  if (display.levels != Display().levels)
    tmp += display_levels_source;
  if (display.soft_clip != Display().soft_clip)
    tmp += display_soft_clip_source;
  
  if (channel)
    tmp += str::Format(channel_source).arg(channel - 1);
  
  return str::Format(shader_source).arg(tmp);
}

double knee(double x, double f)
{
  return math::log(x * f + 1.0) / f;
}

double knee_f(double x, double y)
{
  double f0 = 0.0, f1 = 1.0;

  while (knee(x, f1) > y) {
    f0 = f1;
    f1 = f1 * 2.0;
  }

  for (int i = 0; i < 30; ++i) {
    const double f2 = (f0 + f1) / 2.0;
    if (knee(x, f2) < y) f1 = f2;
    else f0 = f2;
  }

  return (f0 + f1) / 2.0;
}

}

Data::_Shader::_Shader() :
  _profile(profile::PROFILE(0)), _channel(CHANNEL(0)), _init(false),
  _profile_lut(0), _display_lut(0),
  _shader(0), _program(0)
{}

Data::_Shader::~_Shader()
{
  DJV_DBG_OBJ2(_DBG, "Data::_Shader::~_Shader");
  
  if (_profile_lut) DJV_DBG_GL(glDeleteTextures(1, &_profile_lut));
  if (_display_lut) DJV_DBG_GL(glDeleteTextures(1, &_display_lut));
  if (_program) glDeleteObjectARB(_program);
  if (_shader) glDeleteObjectARB(_shader);
}

namespace {

void draw_lut(const image::Data & in, GLuint * out)
{
  DJV_DBG_OBJ2(_DBG, "draw_lut");
  DJV_DBG_PRN("in = " << in);
  
  const image::data::Info & info = in.info();

  const int lut_size = math::to_pow2(info.size.x);
  GLint texture_max = 0;
  DJV_DBG_GL(glGetIntegerv(GL_MAX_TEXTURE_SIZE, &texture_max));
  if (lut_size > texture_max)
    throw Error(error_namespace, str::Format(error_texture_max).
      arg(int(texture_max)));
  DJV_DBG_PRN("lut size = " << lut_size);

  if (*out) { DJV_DBG_GL(glDeleteTextures(1, out)); }
  DJV_DBG_GL(glGenTextures(1, out));
  DJV_DBG_GL(glBindTexture(GL_TEXTURE_1D, *out));
  DJV_DBG_GL(glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_WRAP_S, GL_CLAMP));
  DJV_DBG_GL(glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_MAG_FILTER,
    GL_NEAREST));
  DJV_DBG_GL(glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_MIN_FILTER,
    GL_NEAREST));
  DJV_DBG_GL(glTexImage1D(
    GL_TEXTURE_1D, 0, GL_RGBA,
    lut_size, 0,
    gl::format(info.pixel, info.bgr), gl::type(info.pixel),
    0));
    
  DJV_DBG_GL(glBindTexture(GL_TEXTURE_1D, *out));
  DJV_DBG_GL(glPixelStorei(GL_UNPACK_ALIGNMENT, info.align));
  DJV_DBG_GL(glPixelStorei(GL_UNPACK_SWAP_BYTES,
    info.endian != mem::endian()));
  DJV_DBG_GL(glTexSubImage1D(
    GL_TEXTURE_1D, 0,
    0, lut_size,
    gl::format(info.pixel, info.bgr), gl::type(info.pixel),
    in.data()));
}

}

void Data::_Shader::draw(const Options & in) throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "Data::_Shader::draw");
  
  if (in.profile.type != _profile || in.display != _display ||
    in.channel != _channel
  )
    _init = false;
  _profile = in.profile.type;
  _display = in.display;
  _channel = in.channel;
  
  // Initialize shader.
  
  if (! _init) {
    DJV_DBG_PRN("init");
    
    if (_program) { glDeleteObjectARB(_program); _program = 0; }
    if (_shader) glDeleteObjectARB(_shader);
    DJV_DBG_GL(_shader = glCreateShaderObjectARB(GL_FRAGMENT_SHADER_ARB));
    if (! _shader)
      throw Error(error_namespace, error_shader_create);

    // Shader source code.
    
    const Str tmp = source(_profile, _display, _channel);
    DJV_DBG_PRN("\n" + tmp);
    const char * source_list [] = { tmp.c_str() };
    DJV_DBG_GL(glShaderSourceARB(_shader, 1, source_list, 0));
  
    // Compile shader.
    
    DJV_DBG_GL(glCompileShaderARB(_shader));
    char log [str::cstr_size];
    glGetInfoLogARB(_shader, str::cstr_size, 0, log);
    DJV_DBG_PRN("log = " << log);
    GLint error;
    glGetObjectParameterivARB(_shader, GL_OBJECT_COMPILE_STATUS_ARB, &error);
    if (! error)
      throw Error(error_namespace,
        Str(str::Format(error_shader_compile).arg(log)));

    // Create shader program.
    
    DJV_DBG_GL(_program = glCreateProgramObjectARB());
    if (! _program)
      throw Error(error_namespace, error_shader_program);
    DJV_DBG_GL(glAttachObjectARB(_program, _shader));

    // Link shader program.
    
    DJV_DBG_GL(glLinkProgramARB(_program));
    glGetInfoLogARB(_program, str::cstr_size, 0, log);
    DJV_DBG_PRN("log = " << log);
    glGetObjectParameterivARB(_shader, GL_OBJECT_LINK_STATUS_ARB, &error);
    if (! error)
      throw Error(error_namespace,
        Str(str::Format(error_shader_link).arg(log)));

    _init = true;
  }

  DJV_DBG_GL(glUseProgramObjectARB(_program));
  
  // Shader variables.
  
  if (profile::LUT == _profile) {
    DJV_DBG_GL(glActiveTexture(GL_TEXTURE1));
    draw_lut(in.profile.lut, &_profile_lut);
  }
  
  if (vec::size_valid(_display.lut.size())) {
    DJV_DBG_GL(glActiveTexture(GL_TEXTURE2));
    draw_lut(_display.lut, &_display_lut);
  }
  
  if (_program) {
    DJV_DBG_GL(glUniform1i(glGetUniformLocation(_program, "texture"), 0));
    
    DJV_DBG_GL(glUniform1i(glGetUniformLocation(_program, "profile.lut"), 1));
    
    struct Exposure {
      double v, d, k, f, g;
    } exposure;
    exposure.v = math::pow(2.0, in.profile.exposure.value + 2.47393);
    exposure.d = in.profile.exposure.defog;
    exposure.k = math::pow(2.0, in.profile.exposure.knee_low);
    exposure.f = knee_f(math::pow(2.0, in.profile.exposure.knee_high) -
      exposure.k, math::pow(2.0, 3.5) - exposure.k);
    exposure.g = 1.0 / in.profile.exposure.gamma;
    DJV_DBG_GL(glUniform1f(glGetUniformLocation(_program,
      "profile.exposure.v"), exposure.v));
    DJV_DBG_GL(glUniform1f(glGetUniformLocation(_program,
      "profile.exposure.d"), exposure.d));
    DJV_DBG_GL(glUniform1f(glGetUniformLocation(_program,
      "profile.exposure.k"), exposure.k));
    DJV_DBG_GL(glUniform1f(glGetUniformLocation(_program,
      "profile.exposure.f"), exposure.f));
    DJV_DBG_GL(glUniform1f(glGetUniformLocation(_program,
      "profile.exposure.g"), exposure.g));
    
    DJV_DBG_GL(glUniform1i(glGetUniformLocation(_program, "display.lut"), 2));

    DJV_DBG_GL(glUniformMatrix4fv(glGetUniformLocation(_program,
      "display.color"), 1, false, mtx::transpose(color_mtx(_display.color)).e));

    DJV_DBG_GL(glUniform1f(glGetUniformLocation(_program,
      "display.levels_in"), _display.levels.in[0]));
    DJV_DBG_GL(glUniform1f(glGetUniformLocation(_program,
      "display.levels_in_tmp"), _display.levels.in[1] - _display.levels.in[0]));
    DJV_DBG_GL(glUniform1f(glGetUniformLocation(_program,
      "display.gamma"), 1.0 / _display.levels.gamma));
    DJV_DBG_GL(glUniform1f(glGetUniformLocation(_program,
      "display.levels_out"), _display.levels.out[0]));
    DJV_DBG_GL(glUniform1f(glGetUniformLocation(_program,
      "display.levels_out_tmp"), _display.levels.out[1] -
        _display.levels.out[0]));
    
    DJV_DBG_GL(glUniform1f(glGetUniformLocation(_program,
      "display.soft_clip"), _display.soft_clip));
  }
}

}
}
}
}
