/*------------------------------------------------------------------------------
 lib/djv_image/gl_pbuffer_glx.cxx

 Copyright (c) 2004-2007 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

//#define DJV_DBG
#define _DBG "lib/djv_image/gl_pbuffer_glx.cxx"

#include "gl_pbuffer.h"
#include <djv_base/vec.h>
#include <X11/Xlib.h>
#include <X11/Xatom.h>
#include <X11/Xmu/StdCmap.h>

namespace djv {
namespace image {
namespace gl {
namespace pbuffer {

namespace {

const Str
  error_glew = "Initialize: %1",
  error_glx = "Not Found",
  error_glx_fbconfig = "Frame Buffer Configuration",
  error_glx_pbuffer = "P-Buffer",
  error_glx_context = "Context",
  error_glx_bind = "Context Bind",
  error_x11_display = "Open Display",
  error_x11_visual = "Visual",
  error_x11_colormap = "Colormap",
  error_x11_window = "Window";

class _Pbuffer {
public:

  _Pbuffer(const Vec2i & size, void * context_share) throw (Error);
  ~_Pbuffer();
  
  void current() throw (Error);
  
  const Vec2i & size() const { return _size; }
  void * context_share() const { return _context_share; }
  
private:

  Vec2i _size;
  GLXContext _context;
  void * _context_share;
  GLXPbuffer _pbuffer;
};

class _Pbuffer_Factory {
public:

  _Pbuffer_Factory() throw (Error);
  ~_Pbuffer_Factory();
  
  void add(_Pbuffer *);
  
  _Pbuffer * get(const Vec2i & size, void * context_share) throw (Error);
  
  Display * display() const { return _display; }
  GLXFBConfig config() const { return _config ? _config[_config_id] : 0; }
  
private:

  List<_Pbuffer *> _list;
  Display * _display;
  Colormap _colormap;
  Window _window;
  GLXContext _context;
  GLXFBConfig * _config;
  int _config_size, _config_id;
};

std::auto_ptr<_Pbuffer_Factory> _factory;

_Pbuffer::_Pbuffer(const Vec2i & size, void * context_share) throw (Error) :
  _context(0), _context_share(context_share), _pbuffer(0)
{
  DJV_DBG_OBJ2(_DBG, "_Pbuffer::_Pbuffer");
  DJV_DBG_PRN("size = " << size);
  
  Display * display = _factory->display();
  const GLXFBConfig config = _factory->config();
  if (! display || ! config) return;
  
  _context = glXCreateNewContext(display, config, GLX_RGBA_TYPE,
    (GLXContext)_context_share, True);
  if (! _context) throw Error("GLX", error_glx_context);

  int attrib [] = {
    GLX_PBUFFER_WIDTH, 512, //size.x,
    GLX_PBUFFER_HEIGHT, 512, //size.y,
    GLX_PRESERVED_CONTENTS, True,
    //GLX_LARGEST_PBUFFER, True,
    0
  };
  _pbuffer = glXCreatePbuffer(display, config, attrib);
  if (! _pbuffer) throw Error("GLX", error_glx_pbuffer);
  
  GLuint w = 0, h = 0;
  glXQueryDrawable(display, _pbuffer, GLX_WIDTH_SGIX, &w);
  glXQueryDrawable(display, _pbuffer, GLX_HEIGHT_SGIX, &h);
  _size = Vec2i(w, h);
  DJV_DBG_PRN("real size = " << _size);
}

_Pbuffer::~_Pbuffer()
{
  DJV_DBG_OBJ2(_DBG, "_Pbuffer::~_Pbuffer");

  if (_pbuffer) {
    DJV_DBG_PRN("delete pbuffer");
    glXDestroyPbuffer(_factory->display(), _pbuffer);
  }
  if (_context) {
    DJV_DBG_PRN("delete context");
    glXDestroyContext(_factory->display(), _context);
  }
}

void _Pbuffer::current() throw (Error)
{
  if (! glXMakeCurrent(_factory->display(), _pbuffer, _context))
    throw Error("glx", error_glx_bind);
}

_Pbuffer_Factory::_Pbuffer_Factory() throw (Error) :
  _display(0), _colormap(0), _window(0), _context(0),
  _config(0), _config_size(0), _config_id(0)
{
  DJV_DBG_OBJ2(_DBG, "_Pbuffer_Factory::_Pbuffer_Factory");

  _display = XOpenDisplay(0);
  if (! _display) throw Error("X11", error_x11_display);

  const int screen = DefaultScreen(_display);
  
  if (! glXQueryExtension(_display, 0, 0)) throw Error("GLX", error_glx);

  // XXX Create a dummy window and rendering context for glewInit.
  // According to the docs, glewInit can be called just once per-process?
  
  class Visual {
  public:
  
    Visual(Display * display, int screen) : _p(0)
    {
      int attrib [] = { GLX_RGBA, True, None };
      _p = glXChooseVisual(display, screen, attrib);
    }
    
    ~Visual() { if (_p) XFree(_p); }

    XVisualInfo * operator () () const { return _p; }
  
  private:
  
    XVisualInfo * _p;
  
  } visual(_display, screen);
  
  if (! visual()) throw Error("X11", error_x11_visual);
  //if (visual()->class != TrueColor) throw Error("X11", error_x11_visual);
  
  /*if (XmuLookupStandardColormap(_display, screen, visual()->visualid,
    visual()->depth, XA_RGB_DEFAULT_MAP, False, True)
  ) {
    XStandardColormap * colormap_list = 0;
    int colormap_list_size = 0;
    if (XGetRGBColormaps(_display, RootWindow(_display, screen),
      &colormap_list, &colormap_list_size, XA_RGB_DEFAULT_MAP)
    ) {
      DJV_DBG_PRN("colormaps = " << colormap_list_size);
      for (int i = 0; i < colormap_list_size; ++i)
        if (colormap_list[i].visualid == visual()->visualid) {
          DJV_DBG_PRN("standard colormap = " << i);
          _colormap = colormap_list[i].colormap;
          break;
        }
    }
    XFree(colormap_list);
  }*/
  if (! _colormap) {
    DJV_DBG_PRN("create colormap");
    _colormap = XCreateColormap(_display, RootWindow(_display, screen),
      visual()->visual, AllocNone);
  }
  if (! _colormap) throw Error("X11", error_x11_colormap);
  
  XSetWindowAttributes win_attrib;
  win_attrib.colormap = _colormap;
  win_attrib.border_pixel = 0;
  _window = XCreateWindow(_display, RootWindow(_display, screen),
    0, 0, 1, 1, 0, visual()->depth, InputOutput, visual()->visual,
    CWColormap | CWBorderPixel, &win_attrib);
  if (! _window) throw Error("X11", error_x11_window);

  _context = glXCreateContext(_display, visual(), 0, True);
  if (! _context) throw Error("GLX", error_glx_context);
  
  if (! glXMakeCurrent(_display, _window, _context))
    throw Error("GLX", error_glx_bind);
  
  // Initialize GLEW.
  
  GLint error = glewInit();
  if (error != GLEW_OK)
    throw Error("GLEW", str::Format(error_glew).
      arg((char *)glewGetErrorString(error)));
  
  // Pbuffer configuration.
  
  int fb_attrib [] = {
    GLX_RENDER_TYPE, GLX_RGBA_BIT,
    GLX_DRAWABLE_TYPE, GLX_PBUFFER_BIT_SGIX,
    GLX_DOUBLEBUFFER, False,
    GLX_RED_SIZE, 1,
    GLX_GREEN_SIZE, 1,
    GLX_BLUE_SIZE, 1,
    GLX_ALPHA_SIZE, 0,
    None
  };
  _config = glXChooseFBConfigSGIX(_display, screen, fb_attrib, &_config_size);
  if (! _config || ! _config_size)
    throw Error("GLX", error_glx_fbconfig);
  DJV_DBG_PRN("config size = " << _config_size);
  
  int bit_depth_max [] = { 0, 0, 0, 0 };
  for (int i = 0; i < _config_size; ++i) {
  
    int bit_depth [] = { 0, 0, 0, 0 };
    glXGetFBConfigAttrib(_display, _config[i], GLX_RED_SIZE, &bit_depth[0]);
    glXGetFBConfigAttrib(_display, _config[i], GLX_GREEN_SIZE, &bit_depth[1]);
    glXGetFBConfigAttrib(_display, _config[i], GLX_BLUE_SIZE, &bit_depth[2]);
    glXGetFBConfigAttrib(_display, _config[i], GLX_ALPHA_SIZE, &bit_depth[3]);
    DJV_DBG_PRN("bit depth = " << bit_depth[0] << " " << bit_depth[1] <<
      " " << bit_depth[2] << " " << bit_depth[3]);
    if (
      (bit_depth[0] > bit_depth_max[0]) ||
      (bit_depth[1] > bit_depth_max[1]) ||
      (bit_depth[2] > bit_depth_max[2]) ||
      (bit_depth[3] > bit_depth_max[3])
    ) {
      _config_id = i;
      bit_depth_max[0] = bit_depth[0];
      bit_depth_max[1] = bit_depth[1];
      bit_depth_max[2] = bit_depth[2];
      bit_depth_max[3] = bit_depth[3];
    }
  }
  DJV_DBG_PRN("config = " << bit_depth_max[0] << " " << bit_depth_max[1] <<
    " " << bit_depth_max[2] << " " << bit_depth_max[3]);
}

_Pbuffer_Factory::~_Pbuffer_Factory()
{
  DJV_DBG_OBJ2(_DBG, "_Pbuffer_Factory::~_Pbuffer_Factory");

  for (uint i = 0; i < _list.size(); ++i) delete _list[i];
  if (_config) XFree(_config);
  if (_context) glXDestroyContext(_display, _context);
  if (_window) XDestroyWindow(_display, _window);
  if (_colormap) XFreeColormap(_display, _colormap);
  if (_display) XCloseDisplay(_display);
}

void _Pbuffer_Factory::add(_Pbuffer * in)
{
  DJV_DBG_OBJ2(_DBG, "_Pbuffer_Factory::add");

  if (0) {
    delete in;
  } else {
    _list += in;
    if (int(_list.size()) > 10)
      delete _list.pop_front();
  }
}

_Pbuffer * _Pbuffer_Factory::get(const Vec2i & size, void * context_share)
  throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "_Pbuffer_Factory::get");
  DJV_DBG_PRN("size = " << size);
  
  _Pbuffer * out = 0;

  for (List<_Pbuffer *>::iterator i = _list.begin(); i != _list.end(); ++i)
    if (size <= (*i)->size() && (*i)->context_share() == context_share) {
      out = *i;
      _list.erase(i);
      break;
    }
  
  if (! out) out = new _Pbuffer(size, context_share);
  return out;
}

}

class Pbuffer::Context_Restore {
public:

  Context_Restore() : _context(0), _drawable(0)
  {
    _context = glXGetCurrentContext();
    _drawable = glXGetCurrentDrawable();
  }
  
  ~Context_Restore()
  {
    glXMakeCurrent(_factory->display(), _drawable, _context);
  }
  
private:

  GLXContext _context;
  GLXDrawable _drawable;
};

Pbuffer::Pbuffer(const Vec2i & size, void * context_share) throw (Error) :
  _p(0),
  _context_restore(new Context_Restore)
{
  DJV_DBG_OBJ2(_DBG, "Pbuffer::Pbuffer");
  
  _p = _factory->get(size, context_share);
  _p->current();
}

Pbuffer::~Pbuffer()
{
  DJV_DBG_OBJ2(_DBG, "Pbuffer::~Pbuffer");
  
  if (_p) _factory->add(_p);
}

const Vec2i & Pbuffer::size() const { return _p->size(); }

void init() throw (Error)
  { _factory = std::auto_ptr<_Pbuffer_Factory>(new _Pbuffer_Factory); }
void del() { _factory.reset(); }

}
}
}
}

