/*------------------------------------------------------------------------------
 lib/djv_image/io.cxx

 Copyright (c) 2004-2007 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

//#define DJV_DBG
#define _DBG "lib/djv_image/io.cxx"

#include "io.h"
#include <djv_base/sys.h>
#include <djv_image/io_tga.h>
#include <djv_image/io_sgi.h>
#include <djv_image/io_ppm.h>
#include <djv_image/io_dpx.h>
#include <djv_image/io_cineon.h>
#include <djv_image/io_lut.h>

namespace djv {
namespace image {
namespace io {

Load_Factory::Load_Factory()
{
  DJV_DBG_OBJ2(_DBG, "Load_Factory::Load_Factory");
  
  add(new io::tga::Load);
  add(new io::_sgi::Load);
  add(new io::ppm::Load);
  add(new io::dpx::Load);
  add(new io::cineon::Load);
  add(new io::lut::Load);
  
  set(sys::search_path());
}

Load_Factory::~Load_Factory()
{
  DJV_DBG_OBJ2(_DBG, "Load_Factory::~Load_Factory");
}

namespace {

bool extension_compare(const Str & in, const List<Str> & list)
{
  for (uint i = 0; i < list.size(); ++i)
    if (str::compare(in, list[i])) return true;
  return false;
}

}

bool Load_Factory::valid(const File & in, const Str & plugin)
{
  const List<Load *> & list = this->list();
  for (uint i = 0; i < list.size(); ++i) {    
    if (plugin.size())
      if (! str::compare(plugin, list[i]->name())) continue;
    if (extension_compare(in.extension(), list[i]->extension_list()))
      return true;
  }
  return false;
}

Load * Load_Factory::get(const File & in, Info * info, const Str & plugin)
  throw (Error)
{
  //DJV_DBG_OBJ2(_DBG, "Load_Factory::get");
  //DJV_DBG_PRN("in = " << in);
  //DJV_DBG_PRN("plugin = " << plugin);

  const List<Load *> & list = this->list();
  for (uint i = 0; i < list.size(); ++i) {
    //DJV_DBG_PRN("plugin = " << list[i]->name());
    
    if (plugin.size()) {
      if (! str::compare(plugin, list[i]->name())) continue;
    } else {
      if (! extension_compare(in.extension(), list[i]->extension_list()))
        continue;
    }
    
    std::auto_ptr<Load> plugin(static_cast<Load *>(list[i]->copy()));
    Info info_tmp;
    plugin->open(in, info ? info : &info_tmp);
    return plugin.release();
  }
  
  throw Error(error_namespace, error_unrecognized);
}

bool Load_Factory::option(const Str & plugin, const Str & option,
  List<Str> * data
) {
  DJV_DBG_OBJ2(_DBG, "Load_Factory::option");
  DJV_DBG_PRN("plugin = " << plugin);
  DJV_DBG_PRN("option = " << option);
  DJV_DBG_PRN("data = " << *data);
  
  const List<Load *> & list = this->list();
  for (uint i = 0; i < list.size(); ++i)
    if (str::compare(plugin, list[i]->name())) {
      const List<Str> option_list = list[i]->option_list();
      for (uint j = 0; j < option_list.size(); ++j)
        if (str::compare(option, option_list[j]))
          return list[i]->option(option, data);
    }
  return false;
}

Str Load_Factory::help() const
{
  Str out;
  const List<Load *> & list = this->list();
  for (uint i = 0; i < list.size(); ++i) out += list[i]->help();
  return out;
}

List<Load *> Load_Factory::list() const
{
  const List<plugin::Plugin *> & list = plugin::Factory::list();
  List<Load *> out;
  out.resize(list.size());
  for (uint i = 0; i < list.size(); ++i)
    out[i] = static_cast<Load *>(list[i]);
  return out;
}

Save_Factory::Save_Factory()
{
  DJV_DBG_OBJ2(_DBG, "Save_Factory::Save_Factory");
  
  add(new io::tga::Save);
  add(new io::_sgi::Save);
  add(new io::ppm::Save);
  add(new io::dpx::Save);
  add(new io::cineon::Save);
  add(new io::lut::Save);
  
  set(sys::search_path());
}

Save_Factory::~Save_Factory()
{
  DJV_DBG_OBJ2(_DBG, "Save_Factory::~Save_Factory");
}

Save * Save_Factory::get(const File & in, const Info & info, const Str & plugin)
 throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "Save_Factory::get");
  DJV_DBG_PRN("in = " << in);
  DJV_DBG_PRN("info = " << info);
  DJV_DBG_PRN("plugin = " << plugin);
  
  const List<Save *> & list = this->list();
  for (uint i = 0; i < list.size(); ++i) {
    DJV_DBG_PRN("plugin = " << list[i]->name());
    
    if (plugin.size()) {
      if (! str::compare(plugin, list[i]->name())) continue;
    } else {
      if (! extension_compare(in.extension(), list[i]->extension_list()))
        continue;
    }
    
    std::auto_ptr<Save> plugin(static_cast<Save *>(list[i]->copy()));
    plugin->open(in, info);
    return plugin.release();
  }

  throw Error(error_namespace, error_unrecognized);
}

bool Save_Factory::option(const Str & plugin, const Str & option,
  List<Str> * data
) {
  DJV_DBG_OBJ2(_DBG, "Save_Factory::option");
  DJV_DBG_PRN("plugin = " << plugin);
  DJV_DBG_PRN("option = " << option);
  DJV_DBG_PRN("data = " << *data);

  const List<Save *> & list = this->list();
  for (uint i = 0; i < list.size(); ++i)
    if (str::compare(plugin, list[i]->name())) {
      const List<Str> option_list = list[i]->option_list();
      for (uint j = 0; j < option_list.size(); ++j)
        if (str::compare(option, option_list[j]))
          return list[i]->option(option, data);
    }
  return false;
}

Str Save_Factory::help() const
{
  Str out;
  const List<Save *> & list = this->list();
  for (uint i = 0; i < list.size(); ++i) out += list[i]->help();
  return out;
}

List<Save *> Save_Factory::list() const
{
  const List<plugin::Plugin *> & list = plugin::Factory::list();
  List<Save *> out;
  out.resize(list.size());
  for (uint i = 0; i < list.size(); ++i)
    out[i] = static_cast<Save *>(list[i]);
  return out;
}

namespace {
std::auto_ptr<Load_Factory> _load;
std::auto_ptr<Save_Factory> _save;
}

bool valid(const File & in, const Str & plugin)
{
  return _load->valid(in, plugin);
}

Load * load(const File & in, Info * info, const Str & plugin)
  throw (Error)
{
  Info _info;
  return _load->get(in, info ? info : &_info, plugin);
}

Save * save(const File & in, const Info & info, const Str & plugin)
  throw (Error)
{
  return _save->get(in, info, plugin);
}

Load_Factory * load() { return _load.get(); }
Save_Factory * save() { return _save.get(); }

void init()
{
  DJV_DBG_OBJ2(_DBG, "init");
  
  _load = std::auto_ptr<Load_Factory>(new Load_Factory);
  _save = std::auto_ptr<Save_Factory>(new Save_Factory);
}

void del()
{
  _load.reset();
  _save.reset();
}

}
}
}

