/*------------------------------------------------------------------------------
 lib/djv_image/io_sgi_save.cxx

 Copyright (c) 2004-2007 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

//#define DJV_DBG
#define _DBG "lib/djv_image/io_sgi_save.cxx"

#include "io_sgi.h"

namespace djv {
namespace image {
namespace io {
namespace _sgi {

Save::Save()
{
  _io.endian(mem::endian() != mem::MSB);
}

Str Save::help() const
{
  return str::Format(
"%1:\n"
"* compress (value) - Compression: %2. Default: %3.\n"
  ).
    arg(plugin_name).
    arg(str::join(compress_label(), ", ")).
    arg(compress_label()[_options.compress]);
}

Plugin * Save::copy() const
  { Save * out = new Save; out->_options = _options; return out; }

bool Save::option(const Str & in, List<Str> * data)
{
  if (str::compare(in, option_list()[OPTION_COMPRESS]))
    return str::serialize(data, &_options.compress, compress_label());
  return false;
}

List<Str> Save::option(const Str & in) const
{
  if (str::compare(in, option_list()[OPTION_COMPRESS]))
    return str::serialize(_options.compress, compress_label());
  return List<Str>();
}

const List<Str> & Save::option_list() const { return save_option_label(); }

void Save::open(const File & in, const Info & info) throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "Save::open");
  DJV_DBG_PRN("in = " << in);
  
  _file = in;
  
  // Temporary buffer.
  
  Info _info;
  _info.size = info.size;
  
  pixel::TYPE type = pixel::type(info.pixel);
  switch (type) {
    case pixel::U10:
    case pixel::F16: 
    case pixel::F32: type = pixel::U16; break;
    default: break;
  }
  _info.pixel = pixel::pixel(pixel::format(info.pixel), type);
  
  _info.endian = mem::MSB;

  _image.set(_info);
  DJV_DBG_PRN("image = " << _image);
}

void Save::_open(const File & in) throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "Save::_open");
  DJV_DBG_PRN("in = " << in);

  // Open.

  _io.open(in, djv::io::WRITE);
  
  // Header.
  
  const int channels = _image.channels();
  const int bytes = pixel::channel_bytes(_image.pixel());

  uint16_t magic = 474;
  _io.set_u16(magic);
  
  _io.set_u8(COMPRESS_RLE == _options.compress);
  _io.set_u8(bytes);
  
  const int dimension = 1 == channels ? (1 == _image.h() ? 1 : 2) : 3;
  _io.set_u16(dimension);
  
  _io.set_u16(_image.w());
  _io.set_u16(_image.h());
  _io.set_u16(channels);
  _io.set_u32(0);
  _io.set_u32(1 == bytes ? 255 : 65535);
  _io.position(512);
  
  // Initialize RLE tables.

  if (_options.compress) {
    const int table_size = _image.h() * channels;
    DJV_DBG_PRN("rle table size = " << table_size);
    _rle_offset.size(table_size);
    _rle_size.size(table_size);
    _io.position(table_size * 2 * 4, djv::io::OFFSET);
  }
}

void Save::set(const Image & in, int frame) throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "Save::set");
  DJV_DBG_PRN("in = " << in);
  DJV_DBG_PRN("compress = " << _options.compress);

  // Open.
  
  _open(_file.get(frame));
  
  const Image * image_p = &in;
  if (in.info() != _image.info()) {
    DJV_DBG_PRN("convert = " << _image);
    _image.zero();
    gl::data::copy(in, &_image);
    image_p = &_image;
  }
  
  _image_tmp.set(image_p->info());
  data::planar_deinterleave(*image_p, &_image_tmp);

  // Write.
  
  if (! _options.compress) {

    _io.set(_image_tmp.data(), _image_tmp.data_bytes());

  } else {
  
    const int w = _image_tmp.w(), h = _image_tmp.h();
    const int channels = pixel::channels(_image_tmp.pixel());
    const int bytes = pixel::channel_bytes(_image_tmp.pixel());
    mem::Buffer<uint8_t> scanline(w * bytes * 2);

    for (int c = 0; c < channels; ++c)
      for (int y = 0; y < h; ++y) {
        const size_t size = rle_pack(
          _image_tmp.data() + (c * h + y) * w * bytes, scanline(), w, bytes);
        _rle_offset()[y + c * h] = uint32_t(_io.position());
        _rle_size()[y + c * h] = uint32_t(size);
        _io.set(scanline.get(), size);
      }
  
    _io.position(512);
    _io.set_u32(_rle_offset(), h * channels);
    _io.set_u32(_rle_size(), h * channels);
  }
  
  _io.close();
}

}
}
}
}

