/*------------------------------------------------------------------------------
 lib/djv_image/io_tga_save.cxx

 Copyright (c) 2004-2007 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

//#define DJV_DBG
#define _DBG "lib/djv_image/io_tga_save.cxx"

#include "io_tga.h"

namespace djv {
namespace image {
namespace io {
namespace tga {

Save::Save()
{
  _io.endian(mem::endian() != mem::LSB);
}

Str Save::help() const
{
  return str::Format(
"%1:\n"
"* compress (value) - Compression: %2. Default: %3.\n"
  ).
    arg(plugin_name).
    arg(str::join(compress_label(), ", ")).
    arg(compress_label()[_options.compress]);
}

Plugin * Save::copy() const
  { Save * out = new Save; out->_options = _options; return out; }

bool Save::option(const Str & in, List<Str> * data)
{
  DJV_DBG_OBJ2(_DBG, "Save::option");
  DJV_DBG_PRN("in = " << in);
  DJV_DBG_PRN("data = " << *data);
  
  if (str::compare(in, option_list()[OPTION_COMPRESS]))
    return str::serialize(data, &_options.compress, compress_label());
  return false;
}

List<Str> Save::option(const Str & in) const
{
  DJV_DBG_OBJ2(_DBG, "Save::option");
  DJV_DBG_PRN("in = " << in);
  
  if (str::compare(in, option_list()[OPTION_COMPRESS]))
    return str::serialize(_options.compress, compress_label());
  return List<Str>();
}

const List<Str> & Save::option_list() const { return save_option_label(); }

void Save::open(const File & in, const Info & info) throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "Save::open");
  DJV_DBG_PRN("in = " << in);
  
  _file = in;
  
  // Temporary buffer.
  
  Info _info;
  _info.size = info.size;
  
  _info.pixel = pixel::pixel(pixel::format(info.pixel), pixel::U8);
  switch (pixel::format(_info.pixel)) {
    case pixel::RGB:
    case pixel::RGBA: _info.bgr = true; break;
    default: break;
  }
  
  _info.endian = mem::LSB;
  
  _image.set(_info);
  DJV_DBG_PRN("image = " << _image);
}

void Save::_open(const File & in) throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "Save::_open");
  DJV_DBG_PRN("in = " << in);

  // Open.

  _io.open(in, djv::io::WRITE);
  
  // Header.
  
  uint8_t image_type = 0;
  switch (_image.channels()) {
    case 1:
    case 2: image_type = 3; break;
    case 3:
    case 4: image_type = 2; break;
  }
  if (COMPRESS_RLE == _options.compress) image_type += 8;
  
  int alpha_bits = 0;
  switch (pixel::format(_image.info().pixel)) {
    case pixel::LA:
    case pixel::RGBA: alpha_bits = 8; break;
    default: break;
  }
  const int orient = 0;
  const uint8_t descriptor = (alpha_bits & 15) | ((orient & 3) << 4);

  _io.set_u8(0);
  _io.set_u8(0);
  _io.set_u8(image_type);
  _io.set_u16(0);
  _io.set_u16(0);
  _io.set_u8(0);
  _io.set_u16(0);
  _io.set_u16(0);
  _io.set_u16(_image.w());
  _io.set_u16(_image.h());
  _io.set_u8(_image.channels() * 8);
  _io.set_u8(descriptor);
}

void Save::set(const Image & in, int frame) throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "Save::set");
  DJV_DBG_PRN("in = " << in);

  // Open.
  
  _open(_file.get(frame));
  
  const Image * image_p = &in;
  if (in.info() != _image.info()) {
    DJV_DBG_PRN("convert = " << _image);
    _image.zero();
    gl::data::copy(in, &_image);
    image_p = &_image;
  }
  
  // Write.
  
  if (! _options.compress) {
  
    _io.set(image_p->data(), image_p->data_bytes());
  
  } else {
  
    const int w = image_p->w(), h = image_p->h();
    const int channels = pixel::channels(image_p->info().pixel);
    mem::Buffer<uint8_t> scanline(w * channels * 2);

    for (int y = 0; y < h; ++y) {
      size_t size = rle_pack(image_p->data(0, y), scanline(), w, channels);
      _io.set(scanline.get(), size);
    }
  }

  _io.close();
}

}
}
}
}

