/*------------------------------------------------------------------------------
 lib/djv_ui/prefs_dialog.cxx

 Copyright (c) 2004-2007 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

#include "prefs_dialog.h"
#include <djv_image/io.h>
#include <djv_ui/style.h>
#include <djv_ui/layout_row.h>
#include <djv_ui/group.h>
#include <djv_ui/label.h>
#include <djv_ui/button.h>
#include <djv_ui/choice.h>
#include <djv_ui/tab.h>
#include <djv_ui/image_io.h>

namespace djv {
namespace ui {
namespace prefs_dialog {

Global::Global()
{
  // Widgets.

  group::Box * interface_group = new group::Box("Interface");
  choice::Widget * color_widget = new choice::Widget(style::color_label());
  color_widget->label(new label::Widget("Color:"));
  choice::Widget * size_widget = new choice::Widget(style::size_label());
  size_widget->label(new label::Widget("Size:"));

  group::Box * units_group = new group::Box("Units");
  button::Check * timecode_widget = new button::Check("Timecode");
  
  group::Box * help_group = new group::Box("Help");
  button::Check * tooltips_widget = new button::Check("Tooltips");
  
  // Layout.
  
  layout::group(List<layout::Item *>() << color_widget << size_widget);
  
  layout::Vertical * layout = new layout::Vertical(this);
  layout->spacing(ui::style::style()->spacing_large());
  
  layout->add(interface_group);
  layout::Vertical * group_layout = new layout::Vertical(interface_group);
  group_layout->add(color_widget);
  group_layout->add(size_widget);
  
  layout->add(units_group);
  group_layout = new layout::Vertical(units_group);
  group_layout->add(timecode_widget);
  
  layout->add(help_group);
  group_layout = new layout::Vertical(help_group);
  group_layout->add(tooltips_widget);
  
  layout->add_stretch();
  
  // Initialize.
  
  color_widget->set(style::style()->color());
  size_widget->set(style::style()->size());
  timecode_widget->set(style::style()->timecode());
  tooltips_widget->set(style::style()->tooltips());
  
  // Callbacks.

  color_widget->sig.set(this, color_callback);
  size_widget->sig.set(this, size_callback);
  timecode_widget->sig.set(this, timecode_callback);
  tooltips_widget->sig.set(this, tooltips_callback);
}

void Global::color_callback(int in) { style::style()->color(style::COLOR(in)); }
void Global::size_callback(int in) { style::style()->size(style::SIZE(in)); }
void Global::timecode_callback(bool in) { style::style()->timecode(in); }
void Global::tooltips_callback(bool in) { style::style()->tooltips(in); }

Image_Io::Image_Io()
{
  // Widgets.
  
  tab::Widget * tab_widget = new tab::Widget;
  
  tab::Widget * load_tab_widget = new tab::Widget;
  tab_widget->tab_add("Load", load_tab_widget);
  const List<djv::image::io::Load *> & load_list =
    djv::image::io::load()->list();
  for (uint i = 0; i < load_list.size(); ++i) {
    image_io::Load * widget = image_io::load()->get(load_list[i]);
    if (! widget) continue;
    widget->set(load_list[i]);
    widget->sig.set(this, load_callback);
    load_tab_widget->tab_add(load_list[i]->name(), widget);
  }
  
  tab::Widget * save_tab_widget = new tab::Widget;
  tab_widget->tab_add("Save", save_tab_widget);
  const List<djv::image::io::Save *> & save_list =
    djv::image::io::save()->list();
  for (uint i = 0; i < save_list.size(); ++i) {
    image_io::Save * widget = image_io::save()->get(save_list[i]);
    if (! widget) continue;
    widget->set(save_list[i]);
    widget->sig.set(this, save_callback);
    save_tab_widget->tab_add(save_list[i]->name(), widget);
  }
  
  // Layout.
  
  layout::Vertical * layout = new layout::Vertical(this);
  layout->margin(0);
  layout->add(tab_widget);
}

void Image_Io::load_callback(djv::image::io::Plugin * in)
{
  image_io::sig.emit(in);
}

void Image_Io::save_callback(djv::image::io::Plugin * in)
{
  image_io::sig.emit(in);
}

Dialog::Dialog() :
  dialog::Base("Preferences")
{
  // Widgets.

  _tab_widget = new ui::tab::Widget;
  
  button::Push * close_button = new button::Push("Clos&e");
  
  group::Frame * frame = new group::Frame;
  frame->frame(FL_THIN_DOWN_FRAME);
  
  // Layout.

  layout::Vertical * layout = new layout::Vertical(this);
  
  layout::Vertical * v_layout = new layout::Vertical(layout);
  layout->stretch(v_layout);
  v_layout->margin(0);
  v_layout->add(frame);
  v_layout = new layout::Vertical(frame);
  v_layout->margin(0);
  v_layout->add(_tab_widget);
  
  layout::Horizontal * h_layout = new layout::Horizontal(layout);
  h_layout->margin(0);
  h_layout->add_stretch();
  h_layout->add(close_button);
  
  // Initialize.
  
  size(Vec2i(400, 350));
  
  close_button->take_focus();

  // Callbacks.

  close_button->sig.set(this, close_callback);
}

void Dialog::add(const Str & name, widget::Base * in)
{
  _tab_widget->tab_add(name, in);
}

void Dialog::close_callback(bool)
{
  hide();
}

namespace {
std::auto_ptr<Dialog> _dialog;
}

Dialog * dialog()
{
  if (! _dialog.get()) _dialog = std::auto_ptr<Dialog>(new Dialog);
  return _dialog.get();
}

void del()
{
  if (_dialog.get()) _dialog->widget_del();
  _dialog.release();
}

}
}
}

