/*------------------------------------------------------------------------------
 plugin/djv_exr/save.cxx

 Copyright (c) 2004-2007 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

//#define DJV_DBG
#define _DBG "plugin/djv_exr/save.cxx"

#include "plugin.h"
#include <djv_image/time.h>
#include <ImfChannelList.h>
#include <ImfCompressionAttribute.h>
#include <ImfStandardAttributes.h>
#include <ImfHeader.h>
#include <ImfOutputFile.h>

using namespace image;

extern "C" {
DJV_PLUGIN_EXPORT plugin::Plugin * djv_image_save() { return new Save; }
}

Save::Save() : _f(0) {}
Save::~Save() { _close(); }

Str Save::help() const
{
  return str::Format(
"%1:\n"
"* compress (value) - Compression: %2. Default: %3.\n"
  ).
    arg(plugin_name).
    arg(str::join(compress_label(), ", ")).
    arg(compress_label()[_options.compress]);
}

plugin::Plugin * Save::copy() const
  { Save * out = new Save; out->_options = _options; return out; }

bool Save::option(const Str & in, List<Str> * data)
{
  DJV_DBG_OBJ2(_DBG, "Save::open");
  DJV_DBG_PRN("in = " << in);
  DJV_DBG_PRN("data = " << *data);
  
  DJV_DBG_PRN("compress " << compress_label());
  if (str::compare(in, option_list()[OPTION_COMPRESS]))
    return str::serialize(data, &_options.compress, compress_label());
  return false;
}

List<Str> Save::option(const Str & in) const
{
  if (str::compare(in, option_list()[OPTION_COMPRESS]))
    return str::serialize(_options.compress, compress_label());
  return List<Str>();
}

const List<Str> & Save::option_list() const { return save_option_label(); }

void Save::open(const File & in, const image::io::Info & info) throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "Save::open");
  DJV_DBG_PRN("in = " << in);
  
  _file = in;
  
  // Temporary buffer.
  
  Info _info;
  _info.size = info.size;
  _info.mirror_y = true;
  
  pixel::FORMAT format = pixel::format(info.pixel);
  pixel::TYPE type = pixel::F16;
  switch (type) {
    case pixel::F32: type = pixel::F32; break;
    default: break;
  }
  _info.pixel = pixel::pixel(format, type);
  
  _image.set(_info);
  DJV_DBG_PRN("image = " << _image);
}

void Save::_open(const File & in, const image::io::Info & info) throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "Save::_open");
  DJV_DBG_PRN("in = " << in);

  _close();

try {
  
  // Header.
  
  Imf::Header header(info.size.x, info.size.y);
  
  switch (pixel::channels(info.pixel)) {
    case 1: _channel_list = List<Str>() << "Y"; break;
    case 2: _channel_list = List<Str>() << "Y" << "A"; break;
    case 3: _channel_list = List<Str>() << "R" << "G" << "B"; break;
    case 4: _channel_list = List<Str>() << "R" << "G" << "B" << "A"; break;
  }
  for (uint i = 0; i < _channel_list.size(); ++i)
    header.channels().insert(_channel_list[i].c_str(),
      pixel::type(info.pixel) == pixel::F16 ? Imf::HALF : Imf::FLOAT);
  
  Imf::CompressionAttribute compress;
  switch (_options.compress) {
    case COMPRESS_NONE: compress = Imf::NO_COMPRESSION; break;
    case COMPRESS_PIZ: compress = Imf::PIZ_COMPRESSION; break;
    case COMPRESS_ZIP: compress = Imf::ZIP_COMPRESSION; break;
    case COMPRESS_RLE: compress = Imf::RLE_COMPRESSION; break;
  }
  header.insert(Imf::CompressionAttribute::staticTypeName(), compress);
  
  // Image tags.
  
  Str tag = info.tag[tag::label()[tag::CREATOR]];
  if (tag.size()) addOwner(header, tag);
  tag = info.tag[tag::label()[tag::DESCRIPTION]];
  if (tag.size()) addComments(header, tag);
  tag = info.tag[tag::label()[tag::TIME]];
  if (tag.size()) addCapDate(header, tag);
  tag = info.tag[tag::label()[tag::UTC_OFFSET]];
  if (tag.size()) addutcOffset(header, str::str_to_float(tag));
    
  tag = info.tag[tag_label()[TAG_LONGITUDE]];
  if (tag.size()) addLongitude(header, str::str_to_float(tag));
  tag = info.tag[tag_label()[TAG_LATITUDE]];
  if (tag.size()) addLatitude(header, str::str_to_float(tag));
  tag = info.tag[tag_label()[TAG_ALTITUDE]];
  if (tag.size()) addAltitude(header, str::str_to_float(tag));

  tag = info.tag[tag_label()[TAG_FOCUS]];
  if (tag.size()) addFocus(header, str::str_to_float(tag));
  tag = info.tag[tag_label()[TAG_EXPOSURE]];
  if (tag.size()) addExpTime(header, str::str_to_float(tag));
  tag = info.tag[tag_label()[TAG_APERTURE]];
  if (tag.size()) addAperture(header, str::str_to_float(tag));
  tag = info.tag[tag_label()[TAG_ISO_SPEED]];
  if (tag.size()) addIsoSpeed(header, str::str_to_float(tag));

  tag = info.tag[tag_label()[TAG_CHROMATICITIES]];
  if (tag.size()) {
    List<Str> tmp = str::split(tag, " ");
    if (8 == tmp.size())
      addChromaticities(header, Imf::Chromaticities(
        Imath::V2f(str::str_to_float(tmp[0]), str::str_to_float(tmp[1])),
        Imath::V2f(str::str_to_float(tmp[2]), str::str_to_float(tmp[3])),
        Imath::V2f(str::str_to_float(tmp[4]), str::str_to_float(tmp[5])),
        Imath::V2f(str::str_to_float(tmp[6]), str::str_to_float(tmp[7]))));
  }
  tag = info.tag[tag_label()[TAG_WHITE_LUMINANCE]];
  if (tag.size()) addWhiteLuminance(header, str::str_to_float(tag));
  tag = info.tag[tag_label()[TAG_X_DENSITY]];
  if (tag.size()) addXDensity(header, str::str_to_float(tag));

  tag = info.tag[tag::label()[tag::KEYCODE]];
  if (tag.size()) {
    int id = 0, type = 0, prefix = 0, count = 0, offset = 0;
    image::time::str_to_keycode(tag, &id, &type, &prefix, &count, &offset);
    addKeyCode(header, Imf::KeyCode(id, type, prefix, count, offset));
  }
  tag = info.tag[tag::label()[tag::TIMECODE]];
  if (tag.size()) addTimeCode(header, image::time::str_to_timecode(tag));

  // Open.

  _f = new Imf::OutputFile(in.get().c_str(), header);

} catch (const std::exception & error) {
  throw Error(plugin_name, error.what());
}
}

void Save::_close()
{
  if (_f) { delete _f; _f = 0; }
}

void Save::set(const Image & in, int frame) throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "Save::set");
  DJV_DBG_PRN("in = " << in);
  
  // Open.
  
  image::io::Info info = _image.info();
  info.tag = in.info().tag;
  _open(_file.get(frame), info);
  
  // Temporary buffers.
  
  const Image * image_p = &in;
  if (in.info() != _image.info()) {
    DJV_DBG_PRN("convert = " << _image);
    _image.zero();
    gl::data::copy(in, &_image);
    image_p = &_image;
  }
  
  // Write.

  const int w = image_p->w(), h = image_p->h();
  const int channels = image_p->channels();
  const int bytes = pixel::channel_bytes(image_p->pixel());

try {

  Imf::FrameBuffer frame_buffer;
  for (int c = 0; c < channels; ++c)
    frame_buffer.insert(_channel_list[c].c_str(),
      Imf::Slice(
        pixel::type(image_p->pixel()) == pixel::F16 ? Imf::HALF : Imf::FLOAT,
        (char *)image_p->data() + c * bytes,
        channels * bytes, w * channels * bytes, 1, 1,
        0.0));
  _f->setFrameBuffer(frame_buffer);

  _f->writePixels(h);

} catch (const std::exception & err) {
  throw Error(plugin_name, err.what());
}
  _close();
}
