/*------------------------------------------------------------------------------
 plugin/djv_tiff/save.cxx

 Copyright (c) 2004-2007 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

//#define DJV_DBG
#define _DBG "plugin/djv_tiff/save.cxx"

#include "plugin.h"

using namespace image;

extern "C" {
DJV_PLUGIN_EXPORT plugin::Plugin * djv_image_save()
{
  tiff_init();
  return new Save;
}
}

Save::Save() : _f(0) {}
Save::~Save() { _close(); }

Str Save::help() const
{
  return str::Format(
"%1:\n"
"* compress (value) - Compression: %2. Default: %3.\n"
  ).
    arg(plugin_name).
    arg(str::join(compress_label(), ", ")).
    arg(compress_label()[_options.compress]);
}

plugin::Plugin * Save::copy() const
  { Save * out = new Save; out->_options = _options; return out; }

bool Save::option(const Str & in, List<Str> * data)
{
  if (str::compare(in, option_list()[OPTION_COMPRESS]))
    return str::serialize(data, &_options.compress, compress_label());
  return false;
}

List<Str> Save::option(const Str & in) const
{
  if (str::compare(in, option_list()[OPTION_COMPRESS]))
    return str::serialize(_options.compress, compress_label());
  return List<Str>();
}

const List<Str> & Save::option_list() const { return save_option_label(); }

void Save::open(const File & in, const image::io::Info & info) throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "Save::open");
  DJV_DBG_PRN("in = " << in);
  
  _file = in;
  
  // Temporary buffer.
  
  Info _info;
  _info.size = info.size;
  _info.mirror_y = true;
  
  pixel::TYPE type = pixel::type(info.pixel);
  switch (type) {
    case pixel::U10: type = pixel::U16; break;
    case pixel::F16: type = pixel::F32; break;
    default: break;
  }
  _info.pixel = pixel::pixel(pixel::format(info.pixel), type);
  
  _image.set(_info);
  DJV_DBG_PRN("image = " << _image);
}

void Save::_open(const File & in, const image::tag::Tag & tag) throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "Save::_open");
  DJV_DBG_PRN("in = " << in);

  _close();
  
  // Open.
  
  _f = TIFFOpen(in.get().c_str(), "w");
  if (! _f)
    throw Error(plugin_name, image::io::error_open);
  
  // Header.
  
  uint16 photometric = 0, samples = 0, sample_depth = 0, sample_format = 0;
  uint16 extra_samples [] = { EXTRASAMPLE_ASSOCALPHA };
  uint16 extra_samples_size = 0;
  uint16 compress = 0;
  
  switch (pixel::format(_image.pixel())) {
    case pixel::L:
      photometric = PHOTOMETRIC_MINISBLACK;
      samples = 1;
      break;
    case pixel::LA:
      photometric = PHOTOMETRIC_MINISBLACK;
      samples = 2;
      extra_samples_size = 1;
      break;
    case pixel::RGB:
      photometric = PHOTOMETRIC_RGB;
      samples = 3;
      break;
    case pixel::RGBA:
      photometric = PHOTOMETRIC_RGB;
      samples = 4;
      extra_samples_size = 1;
      break;
  }
  
  switch (pixel::type(_image.pixel())) {
    case pixel::U8:
      sample_depth = 8;
      sample_format = SAMPLEFORMAT_UINT;
      break;
    case pixel::U16:
      sample_depth = 16;
      sample_format = SAMPLEFORMAT_UINT;
      break;
    case pixel::F16:
    case pixel::F32:
      sample_depth = 32;
      sample_format = SAMPLEFORMAT_IEEEFP;
      break;
    default: break;
  }
  
  switch (_options.compress) {
    case COMPRESS_NONE: compress = COMPRESSION_NONE; break;
    case COMPRESS_RLE: compress = COMPRESSION_PACKBITS; break;
    case COMPRESS_LZW: compress = COMPRESSION_LZW; break;
  }
  
  TIFFSetField(_f, TIFFTAG_IMAGEWIDTH, _image.w());
  TIFFSetField(_f, TIFFTAG_IMAGELENGTH, _image.h());
  TIFFSetField(_f, TIFFTAG_PHOTOMETRIC, photometric);
  TIFFSetField(_f, TIFFTAG_SAMPLESPERPIXEL, samples);
  TIFFSetField(_f, TIFFTAG_BITSPERSAMPLE, sample_depth);
  TIFFSetField(_f, TIFFTAG_SAMPLEFORMAT, sample_format);
  TIFFSetField(_f, TIFFTAG_EXTRASAMPLES, extra_samples_size, extra_samples);
  TIFFSetField(_f, TIFFTAG_ORIENTATION, ORIENTATION_TOPLEFT);
  TIFFSetField(_f, TIFFTAG_COMPRESSION, compress);
  TIFFSetField(_f, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG);
  
  // Image tags.
  
  Str tmp = tag[tag::label()[tag::CREATOR]];
  if (tmp.size()) TIFFSetField(_f, TIFFTAG_ARTIST, tmp.c_str());
  tmp = tag[tag::label()[tag::COPYRIGHT]];
  if (tmp.size()) TIFFSetField(_f, TIFFTAG_COPYRIGHT, tmp.c_str());
  tmp = tag[tag::label()[tag::TIME]];
  if (tmp.size()) TIFFSetField(_f, TIFFTAG_DATETIME, tmp.c_str());
  tmp = tag[tag::label()[tag::DESCRIPTION]];
  if (tmp.size()) TIFFSetField(_f, TIFFTAG_IMAGEDESCRIPTION, tmp.c_str());
}

void Save::_close()
{
  if (_f) { TIFFClose(_f); _f = 0; }
}

void Save::set(const Image & in, int frame) throw (Error)
{
  DJV_DBG_OBJ2(_DBG, "Save::set");
  DJV_DBG_PRN("in = " << in);

  // Open.
  
  _open(_file.get(frame), in.info().tag);

  // Temporary buffers.
  
  const Image * image_p = &in;
  if (in.info() != _image.info()) {
    DJV_DBG_PRN("convert = " << _image);
    _image.zero();
    gl::data::copy(in, &_image);
    image_p = &_image;
  }
  
  // Write.
  
  const int h = image_p->h();
  for (int y = 0; y < h; ++y)
    if (TIFFWriteScanline(_f, (tdata_t *)image_p->data(0, y), y) == -1)
      throw Error(plugin_name, image::io::error_write);
  _close();
}

